(********************************************************)
(*   Olivier Boite (under supervision C. Dubois) - 2001 *)
(*   Catherine Dubois - translation to V8 - sept. 2012  *)
(********************************************************)
(*------------------------------------------------------------*)
(*     Generic association table and functions to use it      *)
(*      can be used as a mapping or a partial function        *)
(*      as the domain is explicit                             *)
(*------------------------------------------------------------*)
(* v6.3.1 to v7.2  : Let x:=  is replaced by LetTac x:=       *)
(* and gives a local definition. So Unfold x replaces Heq     *)


Set Implicit Arguments.
Unset Strict Implicit.

Section Generic_Table.
Variable A : Set.

Inductive list : Set :=
  | Nil : list
  | Cons : A -> list -> list.

Variable B : Set.

Inductive table : Set :=
    intro_table : list -> (A -> B) -> table.



Inductive in_prop : A -> list -> Prop :=
  | Hd : forall (x : A) (dom : list), in_prop x (Cons x dom)
  | Tl :
      forall (x x' : A) (dom : list),
      in_prop x dom -> x <> x' -> in_prop x (Cons x' dom).


Definition in_dom (l : A) (t : table) : Prop :=
  match t with
  | intro_table dom s => in_prop l dom
  end.


Definition dom_of (t : table) : list :=
  match t with
  | intro_table d s => d
  end.


(* We require the decidability of the equality on the type A *)
Parameter eq_A_dec : forall x y : A, {x = y} + {x <> y}.



(* (apply_table t a b) stands for t(a)=b and a belongs to dom(t) *)
Inductive apply_table : table -> A -> B -> Prop :=
    intro_apply_table :
      forall (l : A) (dom : list) (f : A -> B),
      in_prop l dom -> apply_table (intro_table dom f) l (f l).


(*---------------------------------*)
Lemma in_dec :
 forall (a : A) (dom : list), {in_prop a dom} + {~ in_prop a dom}.
intros a dom.
induction dom as [| a0 dom Hrecdom].
right.
intro a_in_Nil.
inversion a_in_Nil.

inversion Hrecdom.
case (eq_A_dec a a0).
intro a_is_a0.
left.
rewrite a_is_a0; apply Hd.

intro a_isnt_a; left.
apply Tl.
assumption.

assumption.

case (eq_A_dec a a0).
intro a_is_a0.
left.
rewrite a_is_a0; apply Hd.

intro a_isnt_a.
right.
intro.
inversion H0.
absurd (a = a0); assumption.

absurd (in_prop a dom); assumption.
Qed.

(*---------------------------------*)

(* (add_table a b t) computes the table t' so that    *)
(*   t'(c)=t(c) if c belongs to dom(t)-{a}            *)
(*        =b    if c=a                                *) 

Definition add_table (a : A) (b : B) (t : table) : table :=
  match t with
  | intro_table dom f =>
      match in_dec a dom with
      | left _ =>
          intro_table dom
            (fun a' : A =>
             match eq_A_dec a a' with
             | left _ => b
             | right _ => f a'
             end)
      | right _ =>
          intro_table (Cons a dom)
            (fun a' : A =>
             match eq_A_dec a a' with
             | left _ => b
             | right _ => f a'
             end)
      end
  end.


(*-----------------------------------*)

Definition empty_table (f : A -> B) : table := intro_table Nil f.


(*-----------------------------------*)
Lemma add_dom :
 forall (a a' : A) (b : B) (t : table),
 a <> a' -> in_dom a (add_table a' b t) -> in_dom a t.
intros a a' b t.
elim t.
intros dom f Hneq.
simpl in |- *.
case (in_dec a' dom).
intro a'_in_dom.
simpl in |- *.
intro; assumption.

intro a'_not_in_dom.
simpl in |- *.
intro a_in_cons.
inversion a_in_cons.
absurd (a = a'); assumption.

assumption.

Qed.

(*-----------------------------------*)
Lemma in_dom_add :
 forall (a a' : A) (b : B) (t : table),
 in_dom a t -> in_dom a (add_table a' b t).
intros a a' b t.
elim t.
intros dom f Hin.
simpl in Hin.
simpl in |- *.
case (in_dec a' dom).
intro.
simpl in |- *.
assumption.

intro.
simpl in |- *.
apply Tl.
assumption.

intro a_is_a'.
rewrite a_is_a' in Hin.
absurd (in_prop a' dom); assumption.
Qed.


(*-----------------------------------*)
Lemma add_table_correct :
 forall (t : table) (a : A) (b b' : B),
 apply_table (add_table a b t) a b' -> b = b'.
intros t a b b'.
elim t.
intros dom f.
simpl in |- *.
case (in_dec a dom).
intro a_in_dom.
intro Happ.
inversion Happ.
case (eq_A_dec a a).
intro; trivial.

intro; absurd (a = a); auto.

intros a_not_in_dom Happ.
inversion Happ.
case (eq_A_dec a a).
intro; trivial.

intro; absurd (a = a); auto.

Qed.

(*-----------------------------------*)
Lemma add_table_correct_bis :
 forall (t : table) (a : A) (b : B) (l : list) (f : A -> B),
 intro_table l f = add_table a b t -> b = f a.
intros.
apply add_table_correct with (a := a) (t := t).
rewrite <- H.
apply intro_apply_table.
replace (in_prop a l) with (in_dom a (intro_table l f)).
rewrite H.
unfold add_table in |- *.
case t.
intros.
case (in_dec a l0).
intro; simpl in |- *; assumption.
intro Hnot_in.
simpl in |- *; apply Hd.
trivial.
Qed.


(*-----------------------------------*)
Lemma apply_add :
 forall (t : table) (a : A) (b : B), apply_table (add_table a b t) a b.
intros t a b.
elim t.
intros dom f.
simpl in |- *.
case (in_dec a dom).
intro a_in_dom.
set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b
                | right _ => f a'
                end) in |- *.
 
replace b with (f' a).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a a).
intro; trivial.

intro; absurd (a = a); auto.

intros a_not_in_dom.
set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b
                | right _ => f a'
                end) in |- *.
replace b with (f' a).
apply intro_apply_table.
apply Hd.

unfold f' in |- *.
case (eq_A_dec a a).
intro; trivial.

intro; absurd (a = a); auto.
Qed.




(*-----------------------------------*)
Lemma apply_add_non :
 forall (t : table) (a a' : A) (b b' : B),
 a <> a' -> apply_table t a' b' -> apply_table (add_table a b t) a' b'.
intros t a a' b b' Hneq Happ.
inversion Happ.
simpl in |- *.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ => f a'0
  end) in |- *.
case (in_dec a dom).
intro a_in_dom.
replace (f a') with (f' a').
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a a').
intro; absurd (a = a'); auto.

intro; trivial.

intros a_not_in_dom.
replace (f a') with (f' a').
apply intro_apply_table.
apply Tl.
assumption.

auto.

unfold f' in |- *.
case (eq_A_dec a a').
intro; absurd (a = a'); auto.

intro; trivial.

Qed.

(*-----------------------------------*)


Lemma apply_add_non_bis :
 forall (t : table) (a a' : A) (b b' : B),
 a <> a' -> apply_table (add_table a b t) a' b' -> apply_table t a' b'.

intros t a a' b b' Hneq.
elim t.
intros dom f.
simpl in |- *.
case (in_dec a dom).
intro a_in_dom.
intro Happ.
inversion Happ.
case (eq_A_dec a a').
intro a_is_a'.
absurd (a = a'); auto.

intro a_isnt_a'.
apply intro_apply_table.
assumption.

intros a_not_in_dom.
intro Happ; inversion Happ.
case (eq_A_dec a a').
intro a_is_a'.
absurd (a = a'); auto.

intro a_isnt_a'.
apply intro_apply_table.
inversion H3.
absurd (a = a'); auto.

assumption.
Qed.



(*-------------------------------*)

Lemma swap_table :
 forall (a a' e : A) (b b' r : B) (t : table),
 a <> a' ->
 apply_table (add_table a' b' (add_table a b t)) e r ->
 apply_table (add_table a b (add_table a' b' t)) e r.

intros a a' e b b' r t Hneq.
elim t.
intros dom f.
simpl in |- *.
case (in_dec a dom).
case (in_dec a' dom).
simpl in |- *.
case (in_dec a' dom).
case (in_dec a dom).
intros a_in_dom a'_in_dom a'_in_dom_bis a_in_dom_bis Happ.
inversion Happ.
case (eq_A_dec a' e).
intro a'_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b' with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e).
rewrite <- a'_is_e.
assumption.

assumption.

intro a_isnt_e.
case (eq_A_dec a' e).
intro; trivial.

intro; absurd (a' = e); assumption.

intro a'_isnt_e.
case (eq_A_dec a e).
intro a_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; trivial.

intro; absurd (a = e); assumption.

intro a_isnt_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
assumption.


unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

case (eq_A_dec a' e).
intros; absurd (a' = e); assumption.

intros; trivial.

intros; absurd (in_prop a dom); assumption.

intros; absurd (in_prop a' dom); assumption.

simpl in |- *.
case (in_dec a (Cons a' dom)).
case (in_dec a' dom).
intros; absurd (in_prop a' dom); assumption.

intros a'_not_in_dom a_in_cons_a'_dom a'_not_in_dom_bis a_in_dom.
intros Happ; inversion Happ.
case (eq_A_dec a' e).
intro a'_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b' with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e).
rewrite <- a'_is_e; assumption.

assumption.

case (eq_A_dec a' e).
intros; trivial.

intros; absurd (a' = e); assumption.

intro a'_isnt_e.
case (eq_A_dec a e).
intro a_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intros; trivial.

case (eq_A_dec a' e).
intros; absurd (a = e); assumption.

intros; absurd (a = e); assumption.

intros a_isnt_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

case (eq_A_dec a' e).
intros; absurd (a' = e); assumption.

intros; trivial.

intros a_not_in_cons_a'_dom a'_not_in_dom a_in_dom.
absurd (in_prop a (Cons a' dom)).
assumption.

apply Tl.
assumption.

assumption.

intro a_not_in_dom.
simpl in |- *.
case (in_dec a' (Cons a dom)).
case (in_dec a' dom).
intros a'_in_dom a'_in_cons_a_dom Happ.
inversion Happ.
simpl in |- *.
case (in_dec a dom).
intro a_in_dom.
case (eq_A_dec a' e).
intros a'_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b' with (f' e).
apply intro_apply_table.
inversion H3.
assumption.

assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro a_is_e.
absurd (a = a').
assumption.

rewrite a'_is_e; rewrite a_is_e; trivial.


case (eq_A_dec a' e).
intros; trivial.

intros; absurd (a' = e); assumption.

intros a'_isnt_e.
case (eq_A_dec a e).
intro a_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b with (f' e).
apply intro_apply_table.
rewrite <- a_is_e; assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; trivial.

intro; absurd (a = e); assumption.

intros a_isnt_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
inversion H3.
assumption.

assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

case (eq_A_dec a' e).
intro; absurd (a' = e); assumption.

intros; trivial.

intro a_not_in_dom_bis.
case (eq_A_dec a' e).
intro a'_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b' with (f' e).
apply intro_apply_table.
assumption.


unfold f' in |- *.
case (eq_A_dec a e).
intro a_is_e.
absurd (a = a').
assumption.

rewrite a_is_e; rewrite a'_is_e; trivial.

case (eq_A_dec a' e).
intros; trivial.

intros; absurd (a' = e); assumption.

intros a'_isnt_e.
case (eq_A_dec a e).
intro a_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; trivial.

intro; absurd (a = e); assumption.

intros a_isnt_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

case (eq_A_dec a' e).
intros; absurd (a' = e); assumption.

intros; trivial.

intros a'_not_in_dom a'_in_cons_a_dom.
inversion a'_in_cons_a_dom.
absurd (a = a'); auto.

absurd (in_prop a' dom); assumption.

intros a'_not_in_cons_a_dom Happ.
case (in_dec a' dom).
intro a'_in_dom.
absurd (in_prop a' (Cons a dom)).
assumption.

apply Tl.
assumption.


auto.

intro a'_not_in_dom.
simpl in |- *.
case (in_dec a (Cons a' dom)).
intro a_in_cons_a'_dom.
inversion a_in_cons_a'_dom.
absurd (a = a'); assumption.

absurd (in_prop a dom); assumption.

intro a_not_in_cons_a'_dom.
inversion Happ.
case (eq_A_dec a' e).
intro a'_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b' with (f' e).
apply intro_apply_table.
apply Tl.
rewrite a'_is_e.
apply Hd.

rewrite <- a'_is_e.
auto.

unfold f' in |- *.
case (eq_A_dec a e).
intro a_is_e.
rewrite <- a'_is_e in a_is_e.
absurd (a = a'); assumption.

case (eq_A_dec a' e).
intros; trivial.

intros; absurd (a' = e); assumption.

intros a'_isnt_e.
case (eq_A_dec a e).
intro a_is_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace b with (f' e).
apply intro_apply_table.
rewrite a_is_e; apply Hd.

unfold f' in |- *.
case (eq_A_dec a e).
intro; trivial.

intro; absurd (a = e); assumption.


intro a_isnt_e.
set
 (f' :=
  fun a'0 : A =>
  match eq_A_dec a a'0 with
  | left _ => b
  | right _ =>
      match eq_A_dec a' a'0 with
      | left _ => b'
      | right _ => f a'0
      end
  end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
inversion H3.
apply Tl.
apply Hd.

auto.

inversion H7.
apply Hd.

apply Tl.
apply Tl.
assumption.

assumption.

assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

case (eq_A_dec a' e).
intros; absurd (a' = e); assumption.

intros; trivial.


Qed.


(*-------------------------------*)

Lemma add_add_table :
 forall (t : table) (a e : A) (b b' r : B),
 apply_table (add_table a b t) e r ->
 apply_table (add_table a b (add_table a b' t)) e r.
intros t a e b b' r.
elim t.
intros dom f.
simpl in |- *.


case (in_dec a dom).
intro a_in_dom.
intros Happ; inversion Happ.
case (eq_A_dec a e).
intro a_is_e.
set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b'
                | right _ => f a'
                end) in |- *.
rewrite a_is_e; apply apply_add.

intro a_isnt_e.
apply apply_add_non.
assumption.

set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b'
                | right _ => f a'
                end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

intro; trivial.



intro a_not_in_dom.
intro Happ; inversion Happ.
case (eq_A_dec a e).
intro a_is_e.
set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b'
                | right _ => f a'
                end) in |- *.
rewrite a_is_e; apply apply_add.

intro a_isnt_e.
apply apply_add_non.
assumption.

set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b'
                | right _ => f a'
                end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

intro; trivial.
Qed.



(*-------------------------------*)
Lemma add_add_table_bis :
 forall (t : table) (a e : A) (b b' r : B),
 apply_table (add_table a b (add_table a b' t)) e r ->
 apply_table (add_table a b t) e r.

intros t a e b b' r.
elim t.
intros dom f.
simpl in |- *.
case (in_dec a dom).
simpl in |- *.
case (in_dec a dom).
intros a_in_dom a_in_dom_bis Happ; inversion Happ.
case (eq_A_dec a e).
intro a_is_e.
rewrite a_is_e.
set
 (f' :=
  fun a' : A => match eq_A_dec e a' with
                | left _ => b
                | right _ => f a'
                end) in |- *.
replace b with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec e e).
intro; trivial.

intro; absurd (e = e); auto.

intro a_isnt_e.
set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b
                | right _ => f a'
                end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

intro; trivial.

intros nonP P.
absurd (in_prop a dom); assumption.

intro a_not_in_dom.
simpl in |- *.
case (in_dec a (Cons a dom)).
intro in_cons.
intro Happ.
inversion Happ.
case (eq_A_dec a e).
intro a_is_e.
rewrite a_is_e.
set
 (f' :=
  fun a' : A => match eq_A_dec e a' with
                | left _ => b
                | right _ => f a'
                end) in |- *.
replace b with (f' e).
apply intro_apply_table.
apply Hd.

unfold f' in |- *.
case (eq_A_dec e e).
intro; trivial.

intro; absurd (e = e); auto.

intro a_isnt_e.
set
 (f' :=
  fun a' : A => match eq_A_dec a a' with
                | left _ => b
                | right _ => f a'
                end) in |- *.
replace (f e) with (f' e).
apply intro_apply_table.
assumption.

unfold f' in |- *.
case (eq_A_dec a e).
intro; absurd (a = e); assumption.

intro; trivial.

intro a_not_in_cons.
absurd (in_prop a (Cons a dom)).
assumption.

apply Hd.
Qed.



(*-----------------------------------*)
Inductive equiv : table -> table -> Prop :=
  | eqrefl :
      forall t : table,
      equiv t t
      (*| eqsym  : (t,t':table)(equiv t t')->(equiv t' t) *)
  | eqtrans :
      forall t t' t'' : table, equiv t t' -> equiv t' t'' -> equiv t t''
  | eqhead :
      forall (t t' : table) (a : A) (b : B),
      equiv t t' -> equiv (add_table a b t) (add_table a b t')
  | eqswap :
      forall (t t' : table) (a a' : A) (b b' : B),
      a <> a' ->
      equiv t t' ->
      equiv (add_table a b (add_table a' b' t))
        (add_table a' b' (add_table a b t'))
  | equseless :
      forall (t t' : table) (a : A) (b b' : B),
      equiv (add_table a b t) t' ->
      equiv (add_table a b (add_table a b' t)) t'
  | equseless2 :
      forall (t t' : table) (a : A) (b b' : B),
      equiv t (add_table a b t') ->
      equiv t (add_table a b (add_table a b' t')).

(*-------------------------------*)
Lemma apply_equiv :
 forall (t t' : table) (e : A) (r : B),
 equiv t t' -> apply_table t' e r -> apply_table t e r.
intros t t' e r Hequiv.
induction Hequiv
 as
  [t|
   t t' t'' Hequiv1 HrecHequiv1 Hequiv0 HrecHequiv0|
   t t' a b Hequiv HrecHequiv|
   t t' a a' b b' H Hequiv HrecHequiv|
   t t' a b b' Hequiv HrecHequiv|
   t t' a b b' Hequiv HrecHequiv].

(*refl*)
intro; assumption.

(*trans*)
intro Happ; apply HrecHequiv1; apply HrecHequiv0; assumption.

(*head*)
case (eq_A_dec a e).
intros a_is_e.
rewrite a_is_e.
intros Happ.
cut (b = r).
intros b_is_r.
rewrite b_is_r.
apply apply_add.

apply add_table_correct with (a := e) (t := t').
assumption.

intros a_isnt_e Happ.
apply apply_add_non.
assumption.

apply HrecHequiv.
apply apply_add_non_bis with (a := a) (b := b).
assumption.

assumption.

(*swap*)
case (eq_A_dec a e).
(*a=e*)
intros a_is_e.
rewrite a_is_e.
intros Happ.
cut (b = r).
intros b_is_r.
rewrite b_is_r.
apply apply_add.

apply add_table_correct with (a := e) (t := add_table a' b' t').
apply swap_table.
rewrite <- a_is_e; assumption.
assumption.

(*not a=e*)
intros a_isnt_e.
intros Happ.
apply swap_table.
assumption.
case (eq_A_dec a' e).
intros a'_is_e.
rewrite a'_is_e.
cut (b' = r).
intros b'_is_r.
rewrite b'_is_r.
apply apply_add.
apply add_table_correct with (a := e) (t := add_table a b t').
rewrite a'_is_e in Happ; assumption.
intros a'_isnt_e.
apply apply_add_non.
assumption.
apply apply_add_non.
assumption.
apply HrecHequiv.
apply apply_add_non_bis with (a := a) (b := b).
assumption.
apply apply_add_non_bis with (a := a') (b := b').
assumption.
assumption.

(*useless*)
intros Happ.
apply add_add_table.
apply HrecHequiv; assumption.

(*useless2*)
intros Happ.
apply HrecHequiv.
apply add_add_table_bis with (b' := b').
assumption.
Qed.

(*-------------------------------*)
End Generic_Table.
Set Strict Implicit.
Unset Implicit Arguments.